import uuid

from datetime import date
from typing import Optional

from dirtyfields import DirtyFieldsMixin
from django.conf import settings
from django.db import models
from django.db.models import Lookup, Q
from django.db.models.fields import Field
from django.utils import timezone

from greg.managers import PersonManager


@Field.register_lookup
class Like(Lookup):
    """Allows doing LIKE comparisons when querying."""

    lookup_name = "like"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s LIKE %s" % (lhs, rhs), params


class BaseModel(DirtyFieldsMixin, models.Model):
    """Common fields for all models."""

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class Person(BaseModel):
    """A person is someone who has requested guest access."""

    first_name = models.CharField(max_length=256)
    last_name = models.CharField(max_length=256)
    date_of_birth = models.DateField(null=True)
    registration_completed_date = models.DateField(null=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        related_name="user",
        null=True,
        blank=True,
    )

    objects = PersonManager()

    def __str__(self):
        return "{} {} ({})".format(self.first_name, self.last_name, self.pk)

    def __repr__(self):
        return "{}(id={!r}, first_name={!r}, last_name={!r})".format(
            self.__class__.__name__,
            self.pk,
            self.first_name,
            self.last_name,
        )

    @property
    def private_email(self) -> Optional["Identity"]:
        """The user provided private email address."""
        return self.identities.filter(type=Identity.IdentityType.PRIVATE_EMAIL).first()

    @property
    def private_mobile(self) -> Optional["Identity"]:
        """The user provided private mobile number."""
        return self.identities.filter(
            type=Identity.IdentityType.PRIVATE_MOBILE_NUMBER
        ).first()

    @property
    def fnr(self) -> Optional["Identity"]:
        """The person's fnr if they have one registered."""
        return self.identities.filter(
            type=Identity.IdentityType.NORWEGIAN_NATIONAL_ID_NUMBER
        ).first()

    @property
    def is_registered(self) -> bool:
        """
        A registered guest is a person who has completed
        the registration process.

        The registration process requires that the guest has verified
        their email address via a token link, filled in the required
        personal information along with providing at least one
        verification method, and accepted the institution's mandatory
        consent forms.

        The registered guest may or may not already be verified,
        depending on the verification method.  However, before a
        guest is cleared for account creation at the institution's
        IGA, the guest must be both registered (``is_registered``)
        and verified (``is_verified``).
        """
        # registration_completed_date is set only after accepting consents
        return (
            self.registration_completed_date is not None
            and self.registration_completed_date <= date.today()
        )

    @staticmethod
    def get_verified_identities_query():
        return Q(
            identities__type__in=[
                Identity.IdentityType.NORWEGIAN_NATIONAL_ID_NUMBER,
                Identity.IdentityType.PASSPORT_NUMBER,
            ],
            identities__verified_at__isnull=False,
            identities__verified_at__lte=timezone.now(),
        )

    @property
    def is_verified(self) -> bool:
        """
        A verified guest is a person whom has had their personal
        identity verified.

        Due to the diversity of guests at a university institution,
        there are many ways for guests to identify themselves.
        These include Feide ID, passport number, driver's license,
        national ID card, or another manual (human) verification.

        Some of these methods are implicitly trusted (Feide ID) because
        the guest is likely a visitor from another academic institution
        who has already been pre-verified.  Others are manul, such
        as the sponsor vouching for having checked the guest's
        personal details against his or her passport.

        The verified guest may or may not have completed
        the registration process which implies that it is only
        the combination of being registered (``is_registered``)
        and being verified (``is_verified``) that qualifies for being cleared
        for account creation in the IGA.

        Note that we do not distinguish between the quality,
        authenticity, or trust level of the guest's associated identities.
        """
        return (
            self.identities.filter(
                type__in=[
                    Identity.IdentityType.NORWEGIAN_NATIONAL_ID_NUMBER,
                    Identity.IdentityType.PASSPORT_NUMBER,
                ],
                verified_at__isnull=False,
                verified_at__lte=timezone.now(),
            ).count()
            >= 1
        )

    @property
    def has_mandatory_consents(self) -> bool:
        """
        A person that has mandatory consents is a person whom has given consent
        to all mandatory consent types known to greg.
        """
        person_consents = sorted(
            i.type.name_en for i in self.consents.filter(consent_given_at__isnull=False)
        )
        mandatory_consents = sorted(
            i.name_en for i in ConsentType.objects.filter(mandatory=True)
        )
        return person_consents == mandatory_consents


class RoleType(BaseModel):
    """A role variant."""

    identifier = models.SlugField(max_length=64, unique=True)
    name_nb = models.CharField(max_length=256)
    name_en = models.CharField(max_length=256)
    description_nb = models.TextField()
    description_en = models.TextField()
    default_duration_days = models.IntegerField(null=True)
    # Max days into the future this role can be assigned
    max_days = models.IntegerField(default=365)

    def __str__(self):
        return "{} ({})".format(str(self.name_en or self.name_nb), self.identifier)

    def __repr__(self):
        return "{}(pk={!r}, identifier={!r}, name_nb={!r}, name_en={!r})".format(
            self.__class__.__name__,
            self.pk,
            self.identifier,
            self.name_nb,
            self.name_en,
        )


class Role(BaseModel):
    """The relationship between a person and a role."""

    person = models.ForeignKey("Person", on_delete=models.CASCADE, related_name="roles")
    type = models.ForeignKey(
        "RoleType", on_delete=models.PROTECT, related_name="persons"
    )
    orgunit = models.ForeignKey(
        "OrganizationalUnit", on_delete=models.PROTECT, related_name="unit_person_role"
    )
    # The start date can be null for people that are already
    # attached to the institution but does not have a guest account
    start_date = models.DateField(null=True)
    end_date = models.DateField()
    contact_person_unit = models.TextField(blank=True)
    comments = models.TextField(blank=True)
    available_in_search = models.BooleanField(default=False)
    sponsor = models.ForeignKey(
        "Sponsor", on_delete=models.PROTECT, related_name="sponsor_role"
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["person_id", "type_id", "orgunit_id", "start_date", "end_date"],
                name="person_role_type_unique",
            )
        ]

    def __repr__(self):
        return "{}(id={!r}, person={!r}, type={!r})".format(
            self.__class__.__name__, self.pk, self.person, self.type
        )


class Notification(BaseModel):
    """A change notification that should be delivered to a message queue."""

    identifier = models.IntegerField(null=True, blank=True)
    object_type = models.TextField()
    operation = models.TextField()
    issued_at = models.IntegerField()
    meta = models.JSONField(null=True, blank=True)

    def __repr__(self):
        return "{}(id={!r}, identifier={!r}, object_type={!r}, operation={!r}, issued_at={!r}, meta={!r})".format(
            self.__class__.__name__,
            self.pk,
            self.identifier,
            self.object_type,
            self.operation,
            self.issued_at,
            self.meta,
        )


class Identity(BaseModel):
    class IdentityType(models.TextChoices):
        FEIDE_ID = "feide_id"
        PASSPORT_NUMBER = "passport_number"
        # Norwegian national ID - "fødselsnummer"
        NORWEGIAN_NATIONAL_ID_NUMBER = "norwegian_national_id_number"
        PRIVATE_EMAIL = "private_email"
        PRIVATE_MOBILE_NUMBER = "private_mobile"
        # Sponsor writes what is used in the value column
        OTHER = "other"

    class Verified(models.TextChoices):
        AUTOMATIC = "automatic"
        MANUAL = "manual"

    person = models.ForeignKey(
        "Person", on_delete=models.CASCADE, related_name="identities"
    )
    type = models.CharField(max_length=64, choices=IdentityType.choices)
    source = models.CharField(max_length=256)
    value = models.CharField(max_length=256)
    verified = models.CharField(max_length=9, choices=Verified.choices, blank=True)
    verified_by = models.ForeignKey(
        "Sponsor", on_delete=models.PROTECT, related_name="sponsor", null=True
    )
    verified_at = models.DateTimeField(null=True)

    def __str__(self):
        return "{}(id={!r}, type={!r}, value={!r})".format(
            self.__class__.__name__,
            self.pk,
            self.type,
            self.value,
        )

    def __repr__(self):
        return "{}(id={!r}, person_id={!r}, type={!r}, source={!r}, value={!r}, verified_by={!r}, verified_at={!r})".format(
            self.__class__.__name__,
            self.pk,
            self.person_id,
            self.type,
            self.source,
            self.value,
            self.verified_by,
            self.verified_at,
        )

    class Meta:
        verbose_name_plural = "identities"


class ConsentType(BaseModel):
    """
    Describes some consent, like acknowledging the IT department guidelines, a guest can give.
    """

    identifier = models.SlugField(max_length=64, unique=True)
    name_en = models.CharField(max_length=256)
    name_nb = models.CharField(max_length=256)
    description_en = models.TextField()
    description_nb = models.TextField()
    link_en = models.URLField(null=True)
    link_nb = models.URLField(null=True)
    valid_from = models.DateField(default=date.today)
    user_allowed_to_change = models.BooleanField()
    mandatory = models.BooleanField(default=False)

    def __str__(self):
        return "{} ({})".format(str(self.name_en or self.name_nb), self.identifier)

    def __repr__(self):
        return "{}(id={!r}, identifier={!r}, name_en={!r}, valid_from={!r}, user_allowed_to_change={!r})".format(
            self.__class__.__name__,
            self.pk,
            self.identifier,
            self.name_en,
            self.valid_from,
            self.user_allowed_to_change,
        )


class Consent(BaseModel):
    """
    Links a person and a consent he has given.
    """

    person = models.ForeignKey(
        "Person", on_delete=models.CASCADE, related_name="consents"
    )
    type = models.ForeignKey(
        "ConsentType", on_delete=models.PROTECT, related_name="persons"
    )
    # If the date is blank it means the person has not given consent yet
    consent_given_at = models.DateField(null=True)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["person", "type"], name="person_consent_type_unique"
            )
        ]

    def __repr__(self):
        return "{}(id={!r}, person={!r}, type={!r}, consent_given_at={!r})".format(
            self.__class__.__name__,
            self.pk,
            self.person,
            self.type,
            self.consent_given_at,
        )


class OrganizationalUnit(BaseModel):
    """
    An organizational unit. Units can be organized in a hierarchical manner.
    """

    orgreg_id = models.CharField(max_length=256)
    name_nb = models.CharField(max_length=256)
    name_en = models.CharField(max_length=256)
    parent = models.ForeignKey("self", on_delete=models.PROTECT, null=True, blank=True)
    active = models.BooleanField(default=True)
    deleted = models.BooleanField(default=False)

    def __repr__(self):
        return "{}(id={!r}, orgreg_id={!r}, name_en={!r}, parent={!r})".format(
            self.__class__.__name__, self.pk, self.orgreg_id, self.name_en, self.parent
        )

    def __str__(self):
        return "{} ({})".format(str(self.name_en or self.name_nb), self.orgreg_id)

    class Meta:
        constraints = [
            models.UniqueConstraint(name="unique_orgreg_id", fields=["orgreg_id"])
        ]


class Sponsor(BaseModel):
    """
    A sponsor is someone who is allowed, with some restrictions, to send out invitations to guests and to verify their identity.
    """

    feide_id = models.CharField(max_length=256)
    first_name = models.CharField(max_length=256)
    last_name = models.CharField(max_length=256)

    units = models.ManyToManyField(
        "OrganizationalUnit",
        through="SponsorOrganizationalUnit",
        related_name="sponsor_unit",
    )

    def __str__(self):
        return "{} ({} {})".format(self.feide_id, self.first_name, self.last_name)

    def __repr__(self):
        return "{}(id={!r}, feide_id={!r}, first_name={!r}, last_name={!r})".format(
            self.__class__.__name__,
            self.pk,
            self.feide_id,
            self.first_name,
            self.last_name,
        )

    class Meta:
        constraints = [
            models.UniqueConstraint(name="unique_feide_id", fields=["feide_id"])
        ]


class SponsorOrganizationalUnit(BaseModel):
    """
    A link between a sponsor and an organizational unit.
    """

    sponsor = models.ForeignKey(
        "Sponsor", on_delete=models.PROTECT, related_name="link_sponsor"
    )
    organizational_unit = models.ForeignKey(
        "OrganizationalUnit", on_delete=models.PROTECT, related_name="link_unit"
    )
    hierarchical_access = models.BooleanField()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["sponsor", "organizational_unit"],
                name="sponsor_organizational_unit_unique",
            )
        ]

    def __repr__(self):
        return "{}(id={!r}, sponsor={!r}, organizational_unit={!r}, hierarchical_access={!r})".format(
            self.__class__.__name__,
            self.pk,
            self.sponsor,
            self.organizational_unit,
            self.hierarchical_access,
        )


class ScheduleTask(models.Model):
    """
    Stores information about a task
    """

    name = models.CharField(max_length=32)
    last_completed = models.DateTimeField(null=True)

    def __repr__(self):
        return "{}(id={!r}, name={!r}, last_completed={!r})".format(
            self.__class__.__name__, self.pk, self.name, self.last_completed
        )


class InvitationLink(BaseModel):
    """
    Link to an invitation.

    Having the uuid of an InvitationLink should grant access to the view for posting
    If the Invitation itself is deleted, all InvitationLinks are also be removed.
    """

    uuid = models.UUIDField(null=False, default=uuid.uuid4, blank=False)
    invitation = models.ForeignKey(
        "Invitation", on_delete=models.CASCADE, null=False, blank=False
    )
    expire = models.DateTimeField(blank=False, null=False)


class Invitation(BaseModel):
    """
    Stores information about an invitation.

    Deleting the InvitedPerson deletes the Invitation.
    """

    role = models.ForeignKey("Role", null=False, blank=False, on_delete=models.CASCADE)
