import logging

from rest_framework.authtoken.models import Token
from rest_framework.test import APIClient
from django.contrib.auth import get_user_model
from django.db import transaction

import pytest

from greg.models import (
    Consent,
    Notification,
    Person,
    Sponsor,
    SponsorOrganizationalUnit,
    Identity,
    Role,
    RoleType,
    OrganizationalUnit,
    ConsentType,
)

# faker spams the logs with localisation warnings
# see https://github.com/joke2k/faker/issues/753
logging.getLogger("faker").setLevel(logging.ERROR)


@pytest.fixture
def client() -> APIClient:
    user, _ = get_user_model().objects.get_or_create(username="test")
    token, _ = Token.objects.get_or_create(user=user)
    client = APIClient()
    client.credentials(HTTP_AUTHORIZATION=f"Token {token.key}")
    return client


@pytest.fixture
def pcm_mock():
    class PCMMock:
        def publish(self, *args, **kwargs):
            return True

    return PCMMock()


@pytest.fixture
def person() -> Person:
    pe = Person.objects.create(
        first_name="Test",
        last_name="Tester",
        date_of_birth="2000-01-27",
    )
    return Person.objects.get(id=pe.id)


@pytest.fixture
def person_foo() -> Person:
    with transaction.atomic():
        person = Person.objects.create(
            first_name="Foo",
            last_name="Foo",
            date_of_birth="2001-01-27",
        )
        Identity.objects.create(
            person=person,
            type=Identity.IdentityType.PRIVATE_EMAIL,
            value="foo@example.org",
        )
        Identity.objects.create(
            person=person,
            type=Identity.IdentityType.PRIVATE_MOBILE_NUMBER,
            value="+4712345678",
        )
    return Person.objects.get(id=person.id)


@pytest.fixture
def person_bar() -> Person:
    with transaction.atomic():
        person = Person.objects.create(
            first_name="Bar",
            last_name="Bar",
            date_of_birth="2000-07-01",
        )
        Identity.objects.create(
            person=person,
            type=Identity.IdentityType.PRIVATE_EMAIL,
            value="bar@example.org",
        )
        Identity.objects.create(
            person=person,
            type=Identity.IdentityType.PRIVATE_MOBILE_NUMBER,
            value="+4712345678",
        )
    return Person.objects.get(id=person.id)


@pytest.fixture
def sponsor_guy() -> Sponsor:
    sp = Sponsor.objects.create(
        feide_id="guy@example.org", first_name="Sponsor", last_name="Guy"
    )
    return Sponsor.objects.get(id=sp.id)


@pytest.fixture
def person_foo_verified(person_foo, sponsor_guy) -> Identity:
    pi = Identity.objects.create(
        person=person_foo,
        type=Identity.IdentityType.PASSPORT_NUMBER,
        source="Test",
        value="12345",
        verified=Identity.Verified.MANUAL,
        verified_by=sponsor_guy,
        verified_at="2021-06-15T12:34:56Z",
    )
    return Identity.objects.get(id=pi.id)


@pytest.fixture
def person_foo_not_verified(person_foo) -> Identity:
    pi = Identity.objects.create(
        person=person_foo,
        type=Identity.IdentityType.NORWEGIAN_NATIONAL_ID_NUMBER,
        source="Test",
        value="12345",
    )
    return Identity.objects.get(id=pi.id)


@pytest.fixture()
def role_type_test_guest() -> RoleType:
    rt = RoleType.objects.create(identifier="Test Guest")
    return RoleType.objects.get(id=rt.id)


@pytest.fixture
def unit_foo() -> OrganizationalUnit:
    ou = OrganizationalUnit.objects.create(orgreg_id="12345", name_en="foo_unit")
    return OrganizationalUnit.objects.get(id=ou.id)


@pytest.fixture
def role_person_foo(
    person_foo: Person,
    role_type_test_guest: RoleType,
    sponsor_guy: Sponsor,
    unit_foo: OrganizationalUnit,
) -> Role:
    role = Role.objects.create(
        person=person_foo,
        type=role_type_test_guest,
        start_date="2021-08-02",
        end_date="2021-08-06",
        sponsor=sponsor_guy,
        orgunit=unit_foo,
    )
    return Role.objects.get(id=role.id)


@pytest.fixture
def consent_type() -> ConsentType:
    ct = ConsentType.objects.create(
        identifier="it-guidelines",
        name_en="IT Guidelines",
        name_nb="IT Regelverk",
        description_en="IT Guidelines description",
        description_nb="IT Regelverk beskrivelse",
        link_en="https://example.org/it-guidelines",
        link_nb="https://example.org/it-guidelines",
        user_allowed_to_change=False,
    )
    return ConsentType.objects.get(id=ct.id)


@pytest.fixture
def role_type_foo() -> RoleType:
    rt = RoleType.objects.create(identifier="role_foo", name_en="Role Foo")
    return RoleType.objects.get(id=rt.id)


@pytest.fixture
def role_type_bar() -> RoleType:
    rt = RoleType.objects.create(identifier="role_bar", name_en="Role Bar")
    return RoleType.objects.get(id=rt.id)


@pytest.fixture
def consent_fixture(person, consent_type):
    consent_given_date = "2021-06-20"
    Consent.objects.create(
        person=person, type=consent_type, consent_given_at=consent_given_date
    )
    return Consent.objects.get(id=1)


@pytest.fixture
def consent_type_foo() -> ConsentType:
    return ConsentType.objects.create(
        identifier="test_consent",
        name_en="Test1",
        name_nb="Test2",
        description_en="Test description",
        description_nb="Test beskrivelse",
        link_en="https://example.org",
        link_nb="https://example.org",
        valid_from="2018-01-20",
        user_allowed_to_change=True,
    )


@pytest.fixture
def sponsor_org_unit(sponsor_guy, unit_foo):
    SponsorOrganizationalUnit.objects.create(
        sponsor=sponsor_guy, organizational_unit=unit_foo, hierarchical_access=False
    )
    return SponsorOrganizationalUnit.objects.get(id=1)


@pytest.fixture
def notification():
    Notification.objects.create(
        identifier=5,
        object_type="foo",
        operation="bar",
        issued_at=1234,
        meta="foometa",
    )
    return Notification.objects.get(id=1)
